from datetime import datetime, timezone

from shortuuid import ShortUUID  # type: ignore


def now_epoch() -> int:
    return int(datetime.now(timezone.utc).timestamp())


def get_unique_id() -> str:
    return ShortUUID().random(length=10)



class ShopifyError(Exception):
    """Exception to identify any Shopify error"""

    pass


class ShopifyCallInvalidError(ShopifyError):
    """Exception to identify errors that our due to bad data sent by us

    These should not be retried
    """

    pass


class ShopifyThrottledError(ShopifyError):
    """Exception to identify errors that are due to rate limit control"""

    pass


class ShopifyExceedingMaxCostError(ShopifyError):
    """
    Exception to identify errors that are due to queries exceeding the max query size
    """

    pass


def not_our_fault(exception: Exception):
    """Simple function to identify invalid Shopify calls, i.e. our mistake.

    Probably the only way to make sure we retry for any other exception but those.
    """
    return not isinstance(exception, ShopifyCallInvalidError)



from base64 import b64encode
from hashlib import sha256
from hmac import compare_digest, new


def calculate_from_message(secret: str, message: str, is_base64: bool = False) -> str:
    hmac_hash = new(secret.encode('utf-8'), message.encode('utf-8'), sha256)
    if is_base64:
        # TODO fix bytes / str union issue
        return b64encode(hmac_hash.digest()).decode('utf-8')

    return hmac_hash.hexdigest()


def calculate_from_components(datetime, path, query_string, body, secret, is_base64: bool = False):
    if query_string != '':
        path = path + '?' + query_string
    message = path + datetime + body
    return calculate_from_message(secret=secret, message=message, is_base64=is_base64)


def validate(secret: str, sent_hmac: str, message: str, is_base64: bool = False):

    hmac_calculated = calculate_from_message(secret=secret, message=message, is_base64=is_base64)

    if not compare_digest(sent_hmac, hmac_calculated):
        raise ValueError('HMAC verification failed')



"""
These are the text error codes thrown by the GraphQL API. Although some errors
throw a code e.g.

    "errors": [
        {
            "message": "Throttled",
            "extensions": {
                "code": "THROTTLED"
            }
        }

Not all provide the `code` parameter. Although you could just handle errors
based on the message, this becomes very difficult as some of the messages
are very long (see the max_cost_exceeded error output, which is ~350 characters)
"""

# Error Codes
THROTTLED_ERROR_CODE = 'THROTTLED'
MAX_COST_EXCEEDED_ERROR_CODE = 'MAX_COST_EXCEEDED'
# Error Messages
OPERATION_NAME_REQUIRED_ERROR_MESSAGE = 'An operation name is required'
WRONG_OPERATION_NAME_ERROR_MESSAGE = (
    'No operation named "{}"'  # Required as they don't stay consistant
)



"""A library to facilitate interfacing with Shopify's API"""

__version__ = '0.5.0'


from .token import OfflineTokenABC, OnlineTokenABC, PrivateTokenABC, Token

__all__ = ['OfflineTokenABC', 'OnlineTokenABC', 'PrivateTokenABC', 'Token']



from dataclasses import dataclass
from inspect import isawaitable
from typing import Awaitable, Callable, List, Optional, Union

from fastapi import APIRouter, Depends, HTTPException, Query
from loguru import logger
from starlette.requests import Request
from starlette.responses import RedirectResponse

from ..utils import JWTBaseModel, store_domain
from .redirects import app_redirect, oauth_init_url
from .tokens import OAuthJWT, OfflineToken, OnlineToken
from .validations import validate_callback, validate_oauthjwt


@dataclass
class Callback:
    code: str = Query(...)
    hmac: str = Query(...)
    timestamp: int = Query(...)
    state: str = Query(...)
    shop: str = Query(...)


def init_oauth_router(
    app_scopes: List[str],
    user_scopes: List[str],
    public_domain: str,
    private_key: str,
    app_handle: str,
    api_key: str,
    api_secret_key: str,
    post_install: Callable[[str, OfflineToken], Union[Awaitable[JWTBaseModel], JWTBaseModel]],
    post_login: Optional[Callable[[str, OnlineToken], Optional[Awaitable]]] = None,
    install_init_path='/shopify/auth',
    callback_path='/callback',
    path_prefix: str = '',
) -> APIRouter:
    router = APIRouter()

    if not install_init_path.startswith('/'):
        raise ValueError('The install_init_path argument must start with "/"')
    if not callback_path.startswith('/'):
        raise ValueError('The callback_path argument must start with "/"')
    if path_prefix and not path_prefix.startswith('/'):
        raise ValueError('The path_prefix argument must start with "/"')

    @router.get(install_init_path, include_in_schema=False)
    async def shopify_auth(shop: str):
        """Endpoint initiating the OAuth process on a Shopify store"""
        return RedirectResponse(
            oauth_init_url(
                domain=store_domain(shop=shop),
                is_login=False,
                requested_scopes=app_scopes,
                callback_domain=public_domain,
                callback_path=callback_path,
                path_prefix=path_prefix,
                jwt_key=private_key,
                api_key=api_key,
            )
        )

    @router.get(callback_path, include_in_schema=False)
    async def shopify_callback(request: Request, args: Callback = Depends(Callback)):
        """REST endpoint called by Shopify during the OAuth process for installation and login"""
        try:
            validate_callback(
                shop=args.shop,
                timestamp=args.timestamp,
                query_string=request.scope['query_string'],
                api_secret_key=api_secret_key,
            )
            oauthjwt: OAuthJWT = validate_oauthjwt(
                token=args.state, shop=args.shop, jwt_key=private_key
            )
        except Exception as e:
            raise HTTPException(status_code=400, detail=f'Validation failed: {e}')

        # === If installation ===
        # Setup the login obj and redirect to oauth_redirect
        if not oauthjwt.is_login:
            try:
                # Get the offline token from Shopify
                offline_token = await OfflineToken.get(
                    domain=args.shop,
                    code=args.code,
                    api_key=api_key,
                    api_secret_key=api_secret_key,
                )
            except Exception as e:
                logger.exception(f'Could not retrieve offline token for shop {args.shop}')
                raise HTTPException(status_code=400, detail=str(e))

            # Await if the provided function is async
            if isawaitable(pi_return := post_install(oauthjwt.storename, offline_token)):
                await pi_return  # type: ignore

            if post_login is None:
                return app_redirect(
                    store_domain=args.shop,
                    jwtoken=None,
                    jwt_key=private_key,
                    app_domain=public_domain,
                    app_handle=app_handle,
                )
            # Initiate the oauth loop for login
            return RedirectResponse(
                oauth_init_url(
                    domain=args.shop,
                    is_login=True,
                    requested_scopes=user_scopes,
                    callback_domain=public_domain,
                    callback_path=callback_path,
                    path_prefix=path_prefix,
                    jwt_key=private_key,
                    api_key=api_key,
                )
            )

        # === If login ===
        # Get the online token from Shopify
        online_token = await OnlineToken.get(
            domain=args.shop,
            code=args.code,
            api_key=api_key,
            api_secret_key=api_secret_key,
        )

        # Await if the provided function is async
        jwtoken = None
        if post_login:
            pl_return = post_login(oauthjwt.storename, online_token)
            if isawaitable(pl_return):
                jwtoken = await pl_return  # type: ignore
            else:
                jwtoken = pl_return

        # Redirect to the app in Shopify admin
        return app_redirect(
            store_domain=args.shop,
            jwtoken=jwtoken,
            jwt_key=private_key,
            app_domain=public_domain,
            app_handle=app_handle,
        )

    return router



from typing import List, Optional

from loguru import logger
from pydantic import BaseModel, validator

from ..utils import HTTPClient, JWTBaseModel


class OAuthJWT(JWTBaseModel):
    is_login: bool
    storename: str
    nonce: Optional[str] = None


async def _get_token(
    domain: str,
    code: str,
    api_key: str,
    api_secret_key: str,
) -> dict:
    url = f'https://{domain}/admin/oauth/access_token'

    httpclient = HTTPClient()

    jsondata = {'client_id': api_key, 'client_secret': api_secret_key, 'code': code}
    response = await httpclient.request(
        method='post',
        url=url,
        json=jsondata,
        timeout=20.0,
    )
    if response.status_code != 200:
        message = (
            f'Problem retrieving access token. Status code: {response.status_code} {jsondata}'
            f'response=> {response.text}'
        )
        logger.error(message)
        raise ValueError(message)

    jresp = response.json()

    return jresp


class OfflineToken(BaseModel):
    access_token: str
    scope: List[str]

    @classmethod
    async def get(cls, domain: str, code: str, api_key: str, api_secret_key: str):
        logger.debug(f'Retrieve {cls.__name__} for shop {domain}')
        jsontoken = await _get_token(
            domain=domain,
            code=code,
            api_key=api_key,
            api_secret_key=api_secret_key,
        )
        return cls(**jsontoken)

    @validator('scope', pre=True, always=True)
    def set_id(cls, fld):
        return fld.split(',')


class AssociatedUser(BaseModel):
    id: int
    first_name: str
    last_name: str
    email: str
    email_verified: bool
    account_owner: bool
    locale: str
    collaborator: bool


class OnlineToken(OfflineToken):
    expires_in: int
    associated_user_scope: List[str]
    associated_user: AssociatedUser

    @validator('associated_user_scope', pre=True, always=True)
    def set_id(cls, fld):
        return fld.split(',')



from typing import Optional, Tuple

import jwt
from pydantic import BaseModel, validator

from .misc import now_epoch


class JWTBaseModel(BaseModel):
    """Base class to manage JWT

    The pydantic model fields are the data content of the JWT.
    The default expiration (exp) is set to 900 seconds. Overwrite the ClassVar exp to change the
    expiration time.
    """

    exp: int = None  # type: ignore

    @validator('exp', pre=True, always=True)
    def set_id(cls, exp):
        return exp or (now_epoch() + 900)

    @classmethod
    def decode_token(cls, key: str, token: str, verify: bool = True):
        """Decode the token and load the data content into an instance of this class

        Parameters
        ----------
        key: Secret key used to encrypt the JWT
        verify: If true, verify the signature is valid, otherwise skip. Default is True

        Returns
        -------
        Class instance
        """
        data = jwt.decode(token, key, algorithms=['HS256'], verify=verify)
        # Enforce conversion to satisfy typing
        data = dict(data)
        return cls(**data)

    @classmethod
    def decode_hp_s(cls, key: str, header_payload: str, signature: Optional[str] = None):
        """Decode the token provided in the format "header.payload" and signature

        Parameters
        ----------
        key: Secret key used to encrypt the JWT
        signature: If provided, verify the authenticity of the token.

        Returns
        -------
        Class instance
        """
        sig = signature if signature is not None else 'DUMMYSIGNATURE'
        token = header_payload + '.' + sig
        return cls.decode_token(token=token, key=key, verify=(signature is not None))

    def encode_token(self, key: str) -> str:
        """Encode the class data into a JWT and return a string

        Parameters
        ----------
        key: Secret key used to encrypt the JWT

        Returns
        -------
        The JWT as a string
        """
        data = self.dict()
        data['exp'] = self.exp
        return jwt.encode(data, key, algorithm='HS256')

    def encode_hp_s(self, key: str) -> Tuple[str, str]:
        """Encode the class data into a JWT

        Parameters
        ----------
        key: Secret key used to encrypt the JWT

        Returns
        -------
        The JWT in the format "header.payload" and the signature
        """
        token = self.encode_token(key=key)

        header, payload, signature = token.split('.')
        return f'{header}.{payload}', signature

    @property
    def cookie_key(self) -> str:
        """Define the key of the cookie used to store the JWT"""
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spylib', 'spylib.oauth', 'spylib.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.1.0,<3.0.0',
 'fastapi>=0.69.0',
 'httpx>=0.18.1',
 'loguru>=0.5.3',
 'shortuuid>=1.0.1,<2.0.0',
 'tenacity>=7.0.0,<8.0.0']

setup_kwargs = {
    'name': 'spylib',
    'version': '0.5.0',
    'description': "A library to facilitate interfacing with Shopify's API",
    'long_description': '# Shopify Python Library - SPyLib\n\nThe Shopify python library, or SPyLib, simplifies the use of the Shopify\nservices such as the REST and GraphQL APIs as well as the OAuth authentication.\nAll of this is done **asynchronously using asyncio**.\n\n![Tests](https://github.com/SatelCreative/satel-spylib/actions/workflows/tests.yml/badge.svg)\n\n## Installation\n\n```bash\npip install spylib\n```\n\n## What can SPyLib do?\n\n[Official documentation](https://satelcreative.github.io/spylib)\n\nShopify components managed by SPyLib:\n\n* Admin API\n* Install an app through OAuth\n* Session tokens\n* Webhooks - *Coming soon*\n',
    'author': 'Anthony Hillairet',
    'author_email': 'ant@satel.ca',
    'maintainer': 'Anthony Hillairet',
    'maintainer_email': 'ant@satel.ca',
    'url': 'https://satelcreative.github.io/spylib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
